# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

from dataclasses import dataclass
from typing import Optional

import networkx as nx

from Octeres.bitmask import EnhancedBitmask
from Octeres.bitmask.bitmask_graph_search import bitmask_func_subset
from Octeres.util import superprint


@dataclass
class BitmaskGraphFilterContext:
    """This is used to pass context into the filter function."""

    machine_name: str
    graph: nx.DiGraph
    bitmask_any: Optional[EnhancedBitmask]
    bitmask_all: Optional[EnhancedBitmask]
    bitmask_subset: Optional[EnhancedBitmask]
    node_lst: Optional[list]
    obj_type: Optional[str]
    is_schedulable: Optional[int]

    def filter_func_or_filter_node_names(self, node_name, node_dct, debug=False):
        """match any criteria"""
        if debug:
            superprint(f" *** OR {node_name} {node_dct} *** ")
        _result = False
        if self.bitmask_all is not None:
            _result = node_dct["bitmask"].intersection_all(self.bitmask_all)
        if (not _result) and self.bitmask_any is not None:
            _result = node_dct["bitmask"].intersection_any(self.bitmask_any)
        if (not _result) and self.bitmask_subset is not None:
            _result = bitmask_func_subset(node_name, self.bitmask_subset, node_dct["bitmask"], debug=debug)
        if (not _result) and self.obj_type:
            _result = True if node_dct["obj_type"] == self.obj_type else False
        if (not _result) and self.is_schedulable:
            _result = True if node_dct["is_schedulable"] not in (0, None) else False
        if (not _result) and self.node_lst == ["*"]:
            _result = True
        if (not _result) and self.node_lst:
            _result = True if node_name in self.node_lst else False
        if debug:
            superprint(f"{self=}")
            superprint(f"{node_name=} {_result=} {node_dct=}")

            def noop():
                pass

            (superprint(f'{node_dct["bitmask"].intersection_all(self.bitmask_all)=}') if self.bitmask_all is not None else noop)
            (superprint(f'{node_dct["bitmask"].intersection_any(self.bitmask_any)=}') if self.bitmask_any is not None else noop)
            (
                superprint(f'{bitmask_func_subset(node_name, self.bitmask_subset, node_dct["bitmask"])=}')
                if self.bitmask_subset is not None
                else noop
            )
            (superprint(f'{True if node_dct["obj_type"] == self.obj_type else False=}') if self.obj_type is not None else noop)
            (
                superprint(f'{True if node_dct["is_schedulable"] not in (0, None) else False=}')
                if self.is_schedulable is not None
                else noop
            )
            superprint(f"{True}") if self.node_lst == ["*"] else noop
            (superprint(f"{True if node_name in self.node_lst else False=}") if self.node_lst else noop)
        return _result

    def filter_func_and_filter_node_names(self, node_name, node_dct, debug=False):
        """match all criteria"""
        if debug:
            superprint(f" *** AND {node_name} {node_dct} *** ")
        _result = True
        if self.node_lst != ["*"]:
            if _result and self.obj_type:
                _result = _result & (True if node_dct["obj_type"] == self.obj_type else False)
            if _result and self.is_schedulable:
                _result = _result & (True if node_dct["is_schedulable"] not in (0, None) else False)
            if _result and self.bitmask_all is not None:
                _result = _result & node_dct["bitmask"].intersection_all(self.bitmask_all)
            if _result and self.bitmask_any is not None:
                _result = _result & node_dct["bitmask"].intersection_any(self.bitmask_any)
            if _result and self.bitmask_subset is not None:
                _result = _result & bitmask_func_subset(node_name, self.bitmask_subset, node_dct["bitmask"], debug=debug)
            if _result and self.node_lst:
                _result = _result & (True if node_name in self.node_lst else False)

            if debug:
                superprint(f"{self=}")
                superprint(f"{node_name=} {_result=} {node_dct=}")
                if self.bitmask_all is not None:
                    superprint(f'{node_dct["bitmask"].intersection_all(self.bitmask_all)=}')
                if self.bitmask_any is not None:
                    superprint(f'{node_dct["bitmask"].intersection_any(self.bitmask_any)=}')
                if self.bitmask_subset is not None:
                    superprint(f'{bitmask_func_subset(node_name, self.bitmask_subset, node_dct["bitmask"], debug=debug)}')
                if self.obj_type:
                    superprint(f'{(True if node_dct["obj_type"] == self.obj_type else False)=}')
                if self.is_schedulable:
                    superprint(f'{(True if node_dct["is_schedulable"] not in (0, None) else False)=}')
                if self.node_lst:
                    superprint(f"{(True if node_name in self.node_lst else False)=}")
        else:
            _result = True
        return _result

    def filter_func_subset_obj_type(self, node_name, node_dct, debug=False):
        _result = True
        if _result and self.obj_type:
            _result = _result & (True if node_dct["obj_type"] == self.obj_type else False)
        if _result and self.bitmask_any is not None:
            _result = _result & node_dct["bitmask"].intersection_any(self.bitmask_any)
        if _result and self.bitmask_subset is not None:
            _result = _result & bitmask_func_subset(node_name, self.bitmask_subset, node_dct["bitmask"], debug=debug)
        if debug:
            superprint(f"{self=}")
            superprint(f"{node_name=} {_result=} {node_dct=}")
            if self.bitmask_any is not None:
                superprint(f'{(node_dct["bitmask"] & self.bitmask_any).any()=}')
            if self.bitmask_subset is not None:
                superprint(f'{bitmask_func_subset(node_name, self.bitmask_subset, node_dct["bitmask"], debug=debug)}')
            if self.obj_type:
                superprint(f'{node_dct["obj_type"]=} {self.obj_type=}')
        return _result
