# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

import random
import time
import numpy as np

"""A file containing some random number generation used for non_overlapping_scattered_events"""


class RandomDistribution:
    """Implementing a custom distribution for generating intervals"""

    def __init__(self):
        self.rcache = [random.random() for _ in range(100)]
        self.rcache_idx = 0
        self.cache_size = 100

    def generate_custom_intervals(self):
        if self.rcache_idx >= self.cache_size:
            self.rcache_idx = 0
            self.rcache = [random.random() for _ in range(self.cache_size)]
        r = self.rcache[self.rcache_idx]
        self.rcache_idx += 1

        if r < 0.6:
            result = random.getrandbits(2) + 1  # [1, 4]
        elif r < 0.8:
            result = random.getrandbits(2) + 5  # [5, 8]
        elif r < 0.9:
            result = random.getrandbits(3) + 9  # [9, 16]
        elif r < 0.95:
            result = random.getrandbits(4) + 17  # [17, 32]
        elif r < 0.98:
            result = random.getrandbits(5) + 33  # [33, 64]
        else:
            result = random.getrandbits(6) + 65  # [65, 128]

        return result


class RandomMatrixDistribution:
    """Implementing a random matrix the following the distribution in RandomDistribution.
    Then pick the element from this random matrix would avoid multiple random options."""

    def __init__(self):
        self.distribution = self.generate_distribution()
        self.cache_size = 1000
        self.rcache = np.random.randint(0, len(self.distribution), size=self.cache_size)
        self.rcache_idx = 0

    def generate_distribution(self):
        distribution = []

        distribution.extend([random.getrandbits(2) + 1 for _ in range(480 * 4)])  # [1, 4]
        distribution.extend([random.getrandbits(2) + 5 for _ in range(160 * 4)])  # [5, 8]
        distribution.extend([random.getrandbits(3) + 9 for _ in range(40 * 8)])  # [9, 16]
        distribution.extend([random.getrandbits(4) + 17 for _ in range(10 * 16)])  # [17, 32]
        distribution.extend([random.getrandbits(5) + 33 for _ in range(3 * 32)])  # [33, 64]
        distribution.extend([random.getrandbits(6) + 65 for _ in range(64)])  # [65, 128]

        assert len(distribution) == 3200, "The total length of the distribution does not match 3200."

        np.random.shuffle(distribution)
        return distribution

    def get_random_number(self):
        if self.rcache_idx >= self.cache_size:
            self.rcache_idx = 0
            self.rcache = np.random.randint(0, len(self.distribution), size=self.cache_size)

        r = self.rcache[self.rcache_idx]
        self.rcache_idx += 1

        return self.distribution[r]


class FixedMatrixDistribution:
    """Implementing a fixed matrix in generate_distribution with the same distribution in RandomDistribution.
    This fixed matrix ensure every number in the range of distribution appears at least once in this matrix."""

    def __init__(self):
        self.distribution = self.generate_distribution()
        self.cache_size = 1000
        self.rcache = np.random.randint(0, len(self.distribution), size=self.cache_size)
        self.rcache_idx = 0

    @staticmethod
    def generate_distribution():
        ranges_and_frequencies = [
            (range(1, 5), 480),
            (range(5, 9), 160),
            (range(9, 17), 40),
            (range(17, 33), 10),
            (range(33, 65), 3),
            (range(65, 129), 1),
        ]

        distribution = []
        for num_range, freq in ranges_and_frequencies:
            for num in num_range:
                distribution.extend([num] * freq)

        assert len(distribution) == 3200, "The total length of the distribution does not match 3200."

        np.random.shuffle(distribution)
        return distribution

    def get_random_number(self):
        if self.rcache_idx >= self.cache_size:
            self.rcache_idx = 0
            self.rcache = np.random.randint(0, len(self.distribution), size=self.cache_size)

        r = self.rcache[self.rcache_idx]
        self.rcache_idx += 1

        return self.distribution[r]


def measure_random_distribution():
    """Test the performance of random_distribution in a large range"""
    sr = RandomDistribution()
    start_time = time.time()
    count = 0
    for _ in range(11451419):
        count += sr.generate_custom_intervals()
        if count > 11451419:
            break
    end_time = time.time()
    return end_time - start_time


def measure_random_matrix_distribution():
    """Test the performance of random_matrix_distribution in a large range"""
    ar = RandomMatrixDistribution()
    start_time = time.time()
    count = 0
    for _ in range(11451419):
        count += ar.get_random_number()
        if count > 11451419:
            break
    end_time = time.time()
    return end_time - start_time


def measure_fixed_matrix_distribution():
    """Test the performance of fixed_matrix_distribution in a large range"""
    er = FixedMatrixDistribution()
    start_time = time.time()
    count = 0
    for _ in range(11451419):
        count += er.get_random_number()
        if count > 11451419:
            break
    end_time = time.time()
    return end_time - start_time
