# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""
The Gauntlet
"""

import uuid
import typing
from _collections import defaultdict
from collections import OrderedDict
from datetime import datetime, timedelta
from enum import Enum
from functools import partial
from functools import reduce
from itertools import chain
from itertools import zip_longest
from typing import (
    Optional,
    Tuple,
    List,
    Dict,
    Iterator,
    overload,
    Iterable,
    Union,
    ItemsView,
    ValuesView,
    KeysView,
    Mapping,
    NoReturn,
)
from collections.abc import Callable

import numpy as np
import pandas as pd

from Octeres.bitmask import BASE_DTYPE
from Octeres.bitmask import BitmaskSerializer, EnhancedBitmaskLike
from Octeres.forthwith import FORMAT_DATE_FULL
from Octeres.forthwith import IntervalRange
from Octeres.util import iterate_chain
from typing import TypedDict


T = typing.TypeVar("T")
BitS = BitmaskSerializer()


class DuplicateWarning(Exception):
    pass


def sum_array(a):
    return int(np.sum(a))


class Event_Type(Enum):
    SELECTION = "selection"  # was valid
    NONE = None


class Event_Direction(Enum):
    POSITIVE = 1  # for availability, 1 is up.
    NEGATIVE = -1  # for availability, -1 is down.
    NEUTRAL = 0


class Dependency:
    __slots__ = ["a_pk", "b_pk"]

    def __init__(self, a, b):
        self.a_pk = a["pk"]
        self.b_pk = b["pk"]

    def __repr__(self):
        return "(%s -> %s)" % (self.a_pk, self.b_pk)


class TLEvent(TypedDict):
    """
    something happened in time and space, it has a duration.
    """

    pk: str
    event_type_name: Optional[str]  # this a string because Event_Type can be too many different things.
    bitmask: EnhancedBitmaskLike
    time_start: datetime
    time_end: datetime


class TLPointInTime(TypedDict):
    """
    something happened at this exact point, a change in state.
    """

    pk: str
    name: str
    category: Optional[str]
    bitmask: EnhancedBitmaskLike
    ts: datetime
    direction: Event_Direction


class TLCollision(TypedDict):
    curr_pk: str
    next_pk: str
    intersection_mask: EnhancedBitmaskLike
    intersection_bits: int
    curr_ts: Union[datetime, pd.Timestamp]
    next_ts: Union[datetime, pd.Timestamp]


class TLCollisionOverlap(TLCollision):
    curr_ts_end: Union[datetime, pd.Timestamp]
    overlap_seconds: float


class Point_in_Time(object):
    def __init__(self, base_mask: Optional[EnhancedBitmaskLike] = None):
        # assert sum_array(base_mask) == 0
        self.base_mask = base_mask
        self.positive_lst: List[TLPointInTime] = []  # only use _add functions to append to the list
        self.negative_lst: List[TLPointInTime] = []
        self.neutral_lst: List[TLPointInTime] = []
        self.len_positive_lst = 0
        self.len_negative_lst = 0
        self.len_neutral_lst = 0
        self.positive_mask: Optional[EnhancedBitmaskLike] = None
        self.negative_mask: Optional[EnhancedBitmaskLike] = None
        self.neutral_mask: Optional[EnhancedBitmaskLike] = None

    def __repr__(self, *args, **kwargs):
        return "PIT:+%s:%s:-%s" % (
            len(self.positive_lst),
            len(self.neutral_lst),
            len(self.negative_lst),
        )

    def sum_list_of_masks(self, lst: List[TLPointInTime]) -> EnhancedBitmaskLike:
        bitmask = reduce(lambda a, b: a.add(a, b), [dct["bitmask"] for dct in lst])
        return bitmask

    def get_mask_sum(self) -> EnhancedBitmaskLike:
        final_mask = self.base_mask.copy()
        if self.len_positive_lst > 0:
            final_mask += self.sum_list_of_masks(self.positive_lst)
        if self.len_negative_lst > 0:
            final_mask -= self.sum_list_of_masks(self.negative_lst)
        return final_mask

    def negative_add(self, value_dct: TLPointInTime) -> None:
        self.negative_lst.append(value_dct)
        self.len_negative_lst += 1

    def neutral_add(self, value_dct: TLPointInTime) -> None:
        self.neutral_lst.append(value_dct)
        self.len_neutral_lst += 1

    def positive_add(self, value_dct: TLPointInTime) -> None:
        self.positive_lst.append(value_dct)
        self.len_positive_lst += 1


# List[Tuple[datetime, Point_in_Time]]
class TimelinePit(list):

    def clear(self) -> None:
        super().clear()

    def copy(self) -> List[Tuple[datetime, Point_in_Time]]:
        return super().copy()

    def append(self, __object: Tuple[datetime, Point_in_Time]) -> None:
        super().append(__object)

    def extend(self, __iterable: Iterable[Tuple[datetime, Point_in_Time]]) -> None:
        super().extend(__iterable)

    def pop(self, __index: int = -1) -> Tuple[datetime, Point_in_Time]:
        return super().pop(__index)

    # def index(self, __value: Tuple[datetime, Point_in_Time], __start: int = 0, __stop: int = sys.maxsize) -> int:
    def index(
        self,
        __value: Tuple[datetime, Point_in_Time],
        __start: int = 0,
        __stop: int = 9223372036854775807,
    ) -> int:
        return super().index(__value, __start, __stop)

    def count(self, __value: Tuple[datetime, Point_in_Time]) -> int:
        return super().count(__value)

    def insert(self, __index: int, __object: Tuple[datetime, Point_in_Time]) -> None:
        super().insert(__index, __object)

    def remove(self, __value: Tuple[datetime, Point_in_Time]) -> None:
        super().remove(__value)

    def reverse(self) -> None:
        super().reverse()


# List[Tuple[datetime, EnhancedBitmask]]
class TimelineMask(list):

    def clear(self) -> None:
        super().clear()

    def copy(self) -> List[Tuple[datetime, EnhancedBitmaskLike]]:
        return super().copy()

    def append(self, __object: Tuple[datetime, EnhancedBitmaskLike]) -> None:
        super().append(__object)

    def extend(self, __iterable: Iterable[Tuple[datetime, EnhancedBitmaskLike]]) -> None:
        super().extend(__iterable)

    def pop(self, __index: int = -1) -> Tuple[datetime, EnhancedBitmaskLike]:
        return super().pop(__index)

    # def index(self, __value: Tuple[datetime, EnhancedBitmaskLike], __start: int = 0, __stop: int = sys.maxsize) -> int:
    def index(
        self,
        __value: Tuple[datetime, EnhancedBitmaskLike],
        __start: int = 0,
        __stop: int = 9223372036854775807,
    ) -> int:
        return super().index(__value, __start, __stop)

    def count(self, __value: Tuple[datetime, EnhancedBitmaskLike]) -> int:
        return super().count(__value)

    def insert(self, __index: int, __object: Tuple[datetime, EnhancedBitmaskLike]) -> None:
        super().insert(__index, __object)

    def remove(self, __value: Tuple[datetime, EnhancedBitmaskLike]) -> None:
        super().remove(__value)

    def reverse(self) -> None:
        super().reverse()


# List[Tuple[datetime, int]]
class TimelineMagnitude(list):

    def clear(self) -> None:
        super().clear()

    def copy(self) -> List[Tuple[datetime, int]]:
        return super().copy()

    def append(self, __object: Tuple[datetime, int]) -> None:
        super().append(__object)

    def extend(self, __iterable: Iterable[Tuple[datetime, int]]) -> None:
        super().extend(__iterable)

    def pop(self, __index: int = -1) -> Tuple[datetime, int]:
        return super().pop(__index)

    # def index(self, __value: Tuple[datetime, int], __start: int = 0, __stop: int = sys.maxsize) -> int:
    def index(
        self,
        __value: Tuple[datetime, int],
        __start: int = 0,
        __stop: int = 9223372036854775807,
    ) -> int:
        return super().index(__value, __start, __stop)

    def count(self, __value: Tuple[datetime, int]) -> int:
        return super().count(__value)

    def insert(self, __index: int, __object: Tuple[datetime, int]) -> None:
        super().insert(__index, __object)

    def remove(self, __value: Tuple[datetime, int]) -> None:
        super().remove(__value)

    def reverse(self) -> None:
        super().reverse()


class TimelineDict(dict):
    def clear(self) -> None:
        super().clear()

    def copy(self) -> Dict[datetime, Point_in_Time]:
        return super().copy()

    def popitem(self) -> Tuple[datetime, Point_in_Time]:
        return super().popitem()

    def setdefault(self, __key: datetime, __default: Point_in_Time = None) -> Point_in_Time:
        if __default is None:
            # TODO, figure out a better way.
            raise TypeError("setdefault value: Point_in_Time required for setdefault.")
        return super().setdefault(__key, __default)

    @overload
    def update(self, __m: Mapping[datetime, Point_in_Time], **kwargs: Point_in_Time) -> None: ...

    @overload
    def update(self, __m: Iterable[Tuple[datetime, Point_in_Time]], **kwargs: Point_in_Time) -> None: ...

    @overload
    def update(self, **kwargs: Point_in_Time) -> None: ...

    def update(self, __m: Mapping[datetime, Point_in_Time], **kwargs: Point_in_Time) -> None:
        super().update(__m, **kwargs)

    def keys(self) -> KeysView[datetime]:
        return super().keys()

    def values(self) -> ValuesView[Point_in_Time]:
        return super().values()

    def items(self) -> ItemsView[datetime, Point_in_Time]:
        return super().items()

    @classmethod
    @overload
    def fromkeys(cls, __iterable: Iterable[datetime]) -> Dict[datetime, Point_in_Time]: ...

    @classmethod
    def fromkeys(cls, __iterable: Iterable[datetime], __value: Point_in_Time = None) -> Dict[datetime, Point_in_Time]:
        if __value is None:
            # TODO, figure out a better way.
            raise TypeError("fromkeys value: Point_in_Time required for setdefault.")
        return super().fromkeys(__iterable, __value)


def split_event_set_category(category: Optional[str], event_dct: TLEvent) -> Tuple[TLPointInTime, TLPointInTime]:
    """given an event, split it into start and end
    for now this will mark stuff down as the direction
    is positive."""
    dct_start: TLPointInTime = dict(
        pk=event_dct["pk"],
        name=str(event_dct["pk"]),
        category=category,
        bitmask=event_dct["bitmask"],  # * -1, unit8...
        ts=event_dct["time_start"],
        direction=Event_Direction.POSITIVE,
    )
    dct_end: TLPointInTime = dict(
        pk=event_dct["pk"],
        name=str(event_dct["pk"]),
        category=category,
        bitmask=event_dct["bitmask"],  # * -1, unit8...
        ts=event_dct["time_end"],
        direction=Event_Direction.NEGATIVE,
    )
    return dct_start, dct_end


def split_event(event_dct: TLEvent) -> Tuple[TLPointInTime, TLPointInTime]:
    """given an event, split it into start and end
    for now this will mark stuff down as the direction
    is positive."""
    dct_start: TLPointInTime = dict(
        pk=event_dct["pk"],
        name=str(event_dct["pk"]),
        category=event_dct["event_type_name"],
        bitmask=event_dct["bitmask"],
        ts=event_dct["time_start"],
        direction=Event_Direction.POSITIVE,
    )
    dct_end: TLPointInTime = dict(
        pk=event_dct["pk"],
        name=str(event_dct["pk"]),
        category=event_dct["event_type_name"],
        bitmask=event_dct["bitmask"],
        ts=event_dct["time_end"],
        direction=Event_Direction.NEGATIVE,
    )
    return dct_start, dct_end


def reduce_holes(timeline_holes):
    def _merge_branches(tree):
        new_tree = []
        merged_branches = list()
        for idx in range(0, len(tree), 2):
            left_branch: Tuple[datetime, datetime, EnhancedBitmaskLike] = timeline_holes[idx]
            try:
                right_branch = timeline_holes[idx + 1]
            except IndexError:
                new_tree.append(left_branch)
            else:
                curr_s, curr_e, curr_m = left_branch
                next_s, next_e, next_m = right_branch
                if (curr_m == next_m).all() and curr_e == next_s:
                    # if (curr_m == next_m).all() and curr_e == next_s:
                    new_tree.append((curr_s, next_e, curr_m.copy()))
                    merged_branches.append(idx)
                else:
                    new_tree.append(left_branch)
                    new_tree.append(right_branch)
        return merged_branches, new_tree

    total_reductions = 0
    new_timeline_holes = []
    reductions = -1
    while reductions != 0 and len(timeline_holes) >= 2:
        merged_branches, timeline_holes = _merge_branches(timeline_holes)
        reductions = len(merged_branches)
        total_reductions += reductions
        #
        # since we are comparing pairs, ab cd ef, it's possible that
        # we never compare b to c because a and b equal each other.
        # let's pop off 0 because we know it won't merge, and it will shift the list.
        # Due to the way that this works, it's possible and likely that
        # all timeline points will go through this and the list can be empty
        # thus we must check to make sure we can pop from it, else get out.
        if len(timeline_holes) > 0 and 0 not in merged_branches:
            new_timeline_holes.append(timeline_holes.pop(0))
            # we want to loop again
            reductions += 1
    # lets not forget the remainder
    # this may never do anything except on the last one
    new_timeline_holes.extend(timeline_holes)
    return total_reductions, new_timeline_holes


def _process_timeline(timeline_sorted: List[Tuple[datetime, Point_in_Time]]) -> List[TLCollision]:
    collisions = []
    for idx, curr_ts_pit in enumerate(timeline_sorted):
        curr_ts, curr_pit = curr_ts_pit
        for curr_dct in curr_pit.positive_lst:
            curr_bitmask: EnhancedBitmaskLike = curr_dct["bitmask"]
            curr_pk = curr_dct["pk"]
            stop_iteration = False
            for next_ts_pit in timeline_sorted[idx:]:
                next_ts, next_pit = next_ts_pit
                # let's see if we are done...
                for next_dct in next_pit.negative_lst:
                    next_pk = next_dct["pk"]
                    # lets not compare ourself!
                    if curr_pk == next_pk:
                        stop_iteration = True
                if stop_iteration is False:
                    for next_dct in next_pit.positive_lst:
                        next_bitmask = next_dct["bitmask"]
                        next_pk = next_dct["pk"]
                        # lets not compare ourself!
                        if curr_pk != next_pk:
                            # if intersection_mask is not None and intersection_mask.any():
                            if curr_bitmask.intersection_any(next_bitmask):
                                # print(type(curr_bitmask))
                                intersection_mask = curr_bitmask.intersection(next_bitmask)
                                _curr_ts = curr_ts.to_pydatetime() if isinstance(curr_ts, pd.Timestamp) else curr_ts
                                _next_ts = next_ts.to_pydatetime() if isinstance(next_ts, pd.Timestamp) else next_ts
                                coll_dct: TLCollision = {
                                    "curr_pk": curr_pk,
                                    "next_pk": next_pk,
                                    "intersection_mask": intersection_mask,
                                    "intersection_bits": intersection_mask.bit_count,
                                    "curr_ts": _curr_ts,
                                    "next_ts": _next_ts,
                                }
                                collisions.append(coll_dct)
                else:
                    # we found ourselves
                    break
    return collisions


def _find_overlaps(
    timeline_sorted: List[Tuple[datetime, Point_in_Time]], collisions: List[TLCollision]
) -> List[TLCollisionOverlap]:
    # we need to now find the overlap
    search_for_pk_lst = [dct["curr_pk"] for dct in collisions]
    collisions_with_overlap = []
    pk_lookup = {}
    for ts, pit in timeline_sorted:  # big list, loop first and find
        for pit_dct in pit.negative_lst:
            pit_pk = pit_dct["pk"]
            if pit_pk in search_for_pk_lst:
                pk_lookup[pit_pk] = pit_dct["ts"].to_pydatetime() if isinstance(pit_dct["ts"], pd.Timestamp) else pit_dct["ts"]

    coll_dct: TLCollision
    for coll_dct in collisions:
        curr_pk = coll_dct["curr_pk"]
        next_ts = coll_dct["next_ts"]
        curr_ts_end = pk_lookup[curr_pk]
        overlap_seconds: float = (curr_ts_end - next_ts).total_seconds()

        coll_overlap_dct: TLCollisionOverlap = TLCollisionOverlap(  # type: ignore
            curr_ts_end=curr_ts_end, overlap_seconds=overlap_seconds, **coll_dct
        )
        collisions_with_overlap.append(coll_overlap_dct)
    return collisions_with_overlap


def get_upside_down(number: int) -> str:
    if number == -1:
        result = "⇂"
    elif number == -2:
        result = "ᘔ"
    elif number == -3:
        result = "Ɛ"
    elif number == -4:
        result = "߈"
    elif number == -5:
        result = "ဌ"
    else:
        result = "¡"
    return result


class Timeline(object):
    """a timeline is a well-defined period of time with events.  These
    events can be transformed in to produce MTO graphs, usage, availability
    MTxx, etc.

    For now the class will be treated as a collection of functions that only
    refer to the parameters of the timeline, not the events.  This means no
    references to self.events or such.  Once there is a better understanding
    of how this works, this may change.
    """

    def __init__(
        self,
        machine_name: str,
        range_start: datetime,
        range_end: datetime,
        base_mask: EnhancedBitmaskLike,
    ):
        self.machine_name: str = machine_name
        self.range_start: datetime = range_start
        self.range_end: datetime = range_end
        assert sum_array(base_mask) == 0
        self.base_mask: EnhancedBitmaskLike = base_mask
        self.bitmask_class = base_mask.__class__
        self.possible_unit_seconds: float = (range_end - range_start).total_seconds() * len(base_mask)

    def get_possible_unit_seconds(self) -> float:
        return self.possible_unit_seconds

    def generate_selection_events(self) -> List[TLPointInTime]:
        ts = self.range_start
        te = self.range_end
        machine_name = self.machine_name
        base_mask = self.base_mask
        timeline = []
        pk = uuid.uuid1().hex
        dct_start: TLPointInTime = dict(
            pk=pk,
            name=machine_name,
            category=Event_Type.SELECTION.value,
            bitmask=base_mask,  # this needs to be all zeros!
            ts=ts,
            direction=Event_Direction.NEUTRAL,
        )
        dct_end: TLPointInTime = dict(
            pk=pk,
            name=machine_name,
            category=Event_Type.SELECTION.value,
            bitmask=base_mask,  # this needs to be all zeros!
            ts=te,
            direction=Event_Direction.NEUTRAL,
        )
        timeline.append(dct_start)
        timeline.append(dct_end)
        return timeline

    def _filter_events(self, range_start: datetime, range_end: datetime, lst_of_dct: List[TLEvent]) -> List[TLEvent]:
        def filter_func(dct):
            return True if dct["time_start"] <= range_end and dct["time_end"] > range_start else False

        lst_of_dct = list(filter(filter_func, lst_of_dct))
        return lst_of_dct

    def split_events(self, lst_of_dct: List[TLEvent], category=None) -> Iterator[TLPointInTime]:
        timeline: Iterator[TLPointInTime]
        if category:
            func = partial(split_event_set_category, category)
            timeline = chain.from_iterable(map(func, lst_of_dct))
        else:
            timeline = chain.from_iterable(map(split_event, lst_of_dct))
        return timeline

    def prepare_event_timeline(
        self,
        lst_of_dct: List[TLEvent],
        category: Optional[str] = None,
        noselection=False,
    ) -> List[TLPointInTime]:
        """create a timeline of the events"""
        lst_of_dct = self._filter_events(self.range_start, self.range_end, lst_of_dct)
        timeline: List[TLPointInTime] = []
        if not noselection:
            timeline.extend(self.generate_selection_events())
        # only add the events if they intersect the range.
        if category:
            timeline.extend(self.split_events(lst_of_dct, category))
        else:
            timeline.extend(self.split_events(lst_of_dct))
        return timeline

    def group_timeline(self, timeline: List[TLPointInTime]) -> TimelineDict:
        """allow for order of the timeline by creating a dictionary with the keys
        as timestamps and a list of half_events with the ends first.

        this prevents _____|_____ or toggling for the same second.
        """
        # now sort the timeline by ts, if two are the same, ends go first
        timeline_dct = TimelineDict()
        ts_dct: TLPointInTime
        for ts_dct in timeline:
            direction = ts_dct["direction"]
            ts = ts_dct["ts"]
            try:
                timeline_dct[ts]
            except KeyError:
                timeline_dct[ts] = Point_in_Time(self.base_mask)
            # need to order by ends first, we we might need to!

            if direction == Event_Direction.POSITIVE:
                timeline_dct[ts].positive_add(ts_dct)
            elif direction == Event_Direction.NEGATIVE:
                timeline_dct[ts].negative_add(ts_dct)
            else:  # Event_Direction.NEUTRAL
                timeline_dct[ts].neutral_add(ts_dct)
        return timeline_dct

    def sort_timeline(self, timeline_dct: Dict[datetime, Point_in_Time]) -> TimelinePit:
        """sorts a grouped timeline."""
        timeline_sorted: List[Tuple[datetime, Point_in_Time]] = []
        for ts in sorted(timeline_dct.keys()):
            pit: Point_in_Time = timeline_dct[ts]
            timeline_sorted.append((ts, pit))
        return TimelinePit(timeline_sorted)

    def find_collisions_timeline(self, timeline_sorted: TimelinePit) -> List[TLCollisionOverlap]:
        """
        This is similar to normalize_timeline except that it compares all the
        points in time with each other to find any intersections.
        """
        collisions = _process_timeline(timeline_sorted)
        collisions_with_overlap = _find_overlaps(timeline_sorted, collisions)
        return collisions_with_overlap

    def normalize_timeline_pit(self, timeline_pit: TimelinePit) -> TimelineMask:
        mask_timeline = TimelineMask()
        ts: datetime
        pit: Point_in_Time
        for (ts, pit), first in zip_longest(timeline_pit, [True], fillvalue=False):
            timeline_mask: EnhancedBitmaskLike = pit.get_mask_sum()
            mask_timeline.append((ts, timeline_mask))
        return mask_timeline

    def flatten_timeline(self, mask_timeline: TimelineMask) -> TimelineMask:
        """this will just take the sign of the timeline."""
        flattened_mask_timeline = TimelineMask()
        ts: datetime
        pit: Point_in_Time
        for ts, bitmask in mask_timeline:
            flattened_mask_timeline.append((ts, bitmask.sign(bitmask)))
        return flattened_mask_timeline

    def sum_timeline(self, mask_timeline: TimelineMask, test_negative=False) -> TimelineMask:
        """walk through the timeline and combine the masks with the previous"""
        sum_mask_timeline = TimelineMask()
        ts: datetime
        pit: Point_in_Time
        for (ts, bitmask), first in zip_longest(mask_timeline, [True], fillvalue=False):
            base = bitmask.copy() if first else base + bitmask  # base will be defined due to first
            if test_negative:
                if (base < 0).any():
                    raise Exception("TimelineMask contains a negative number.")
            sum_mask_timeline.append((ts, base))
        return sum_mask_timeline

    def normalize_timeline(self, timeline_sorted: TimelinePit, flatten=True, test_negative=False) -> TimelineMask:
        """
        This does 3 things.
        1.  combines all the masks for each timestamp(ts)
        2.  creates a new mask walking forward by adding (+)
        3.  reduces any collisions(>1 in the mask) to +1,0,-1 using np.sign
        """
        mask_timeline = self.normalize_timeline_pit(timeline_sorted)
        if flatten:
            mask_timeline = self.sum_timeline(mask_timeline, test_negative=test_negative)
            mask_timeline = self.flatten_timeline(mask_timeline)
        else:
            mask_timeline = self.sum_timeline(mask_timeline)
        return mask_timeline

    def isolate_timeline_range(self, mask_timeline: TimelineMask) -> TimelineMask:
        """looking for the selection, walk the timeline and find it and cut all the events."""
        range_start = self.range_start
        range_end = self.range_end
        in_range = False
        new_mask_timeline = TimelineMask()
        for ts, sign_mask in mask_timeline:
            if ts >= range_start:
                in_range = True
            if ts > range_end:
                in_range = False
            if in_range is True:
                new_mask_timeline.append((ts, sign_mask))
        return new_mask_timeline

    def timeline_apply_function(self, timeline, func):
        """this will apply a function to the timeline"""
        vfunc = np.vectorize(func)
        timeline_new = []
        for ts, bitmask in timeline:
            new_bitmask = np.apply_along_axis(vfunc, 0, bitmask)  # FIXME: priority 1 numpy
            timeline_new.append((ts, new_bitmask))
        return timeline_new

    def print_timeline(self, timeline, binary=False, name=""):
        print(f"-----v-v-----  printing timeline {name} -----v-v-----")
        for ts, element in timeline:
            if type(element) == Point_in_Time:
                print("%-2s %-2s" % (ts.strftime(FORMAT_DATE_FULL), element))
            elif isinstance(element, EnhancedBitmaskLike):
                if binary:
                    other_str = "".join([str(_) if _ >= 0 else get_upside_down(_) for _ in element])  # FIXME: priority 1 iterable
                    print("%s %6s %s" % (ts.strftime(FORMAT_DATE_FULL), sum_array(element), other_str))
                else:
                    try:
                        hexmask = BitS.encode_bitmask_to_hex(element)
                        print(
                            "%s %6s %s"
                            % (
                                ts.strftime(FORMAT_DATE_FULL),
                                sum_array(element),
                                hexmask,
                            )
                        )
                    except KeyError:
                        # negative bitmasks
                        other_str = "binary:" + "".join(
                            [str(_) if _ >= 0 else get_upside_down(_) for _ in element]
                        )  # FIXME: priority 1 iterable
                        print(
                            "%s %6s %s"
                            % (
                                ts.strftime(FORMAT_DATE_FULL),
                                sum_array(element),
                                other_str,
                            )
                        )
            else:
                print("%s %6s %s" % (ts.strftime(FORMAT_DATE_FULL), sum_array(element), element))
        print(f"-----^-^-----  printed timeline {name} -----^-^-----")

    def add_edges(self, timeline_mag: TimelineMagnitude) -> TimelineMagnitude:
        """this will add edges to the timeline.
        If usage was 20 for a long time and then jumped to 30, you would see an
        angled line even though the integral for the line should be much lower.
        Add a 20 ts at the start of thirty and then the same ts at 30, in order.
        This will create a more blocky graph.
        #     /---           |-------
        #    /      to       |
        #???/             ___|
        #   a b            a b
        """
        edged_timeline = TimelineMagnitude()
        # now we need to walk time, clean up start, cleanup end, cut events
        ts: datetime
        magnitude_previous: int
        ts, magnitude_previous = timeline_mag[0]
        # if array, might need to use copy(), if int, we are ok.
        edged_timeline.append((ts, magnitude_previous))
        # edged_timeline: List[Tuple[datetime, int]] = [(ts, magnitude_previous)]
        # print(ts, magnitude_previous)
        for ts, value in timeline_mag[1:]:
            delta = value - magnitude_previous
            magnitude: int = magnitude_previous + delta
            # print("TS:%2s Value:%2s, Delta:%2s Prev:%s Curr:%s" % (ts, value, delta, magnitude_previous, magnitude))
            edged_timeline.append((ts, magnitude_previous))
            edged_timeline.append((ts, magnitude))
            magnitude_previous = magnitude
        return edged_timeline

    def mask_timeline_to_mag_timeline(self, mask_timeline: TimelineMask) -> TimelineMagnitude:
        """Converts a mask_timeline with masks to a magnitude mask_timeline
        this assumes that the mask_timeline is already binary for the sum.
        """
        # first sum all the masks up to get magnitudes
        mag_timeline = TimelineMagnitude()
        ts: datetime
        bitmask: EnhancedBitmaskLike
        for ts, bitmask in mask_timeline:
            mag_timeline.append((ts, sum(bitmask)))
        # now add edges or the graph will look like a saw.
        # FIXME: we need to move this out
        mag_timeline = self.add_edges(mag_timeline)
        return mag_timeline

    def search_timeline_for_holes(self, mask_timeline: TimelineMask):
        """this will walk forward in time to try to find any
        non-set bits.  This will aid in detecting and isolating
        missing data"""
        timeline_holes = []
        base_mask: Optional[EnhancedBitmaskLike] = None
        for prev_ts_bitmask, curr_ts_bitmask, next_ts_bitmask in iterate_chain(mask_timeline):
            curr_ts, curr_bitmask = curr_ts_bitmask
            # assert curr_bitmask.dtype == BASE_DTYPE
            if prev_ts_bitmask is None:
                # we found the start!
                base_mask = self.bitmask_class.ones(len(curr_bitmask))
            if next_ts_bitmask is not None:
                next_ts, next_bitmask = next_ts_bitmask
                delta_mask = curr_bitmask ^ base_mask
                if delta_mask.any():
                    timeline_holes.append((curr_ts, next_ts, delta_mask))

        reductions, timeline_holes = reduce_holes(timeline_holes)
        return timeline_holes

    def calculate_area(
        self,
        mag_timeline: TimelineMagnitude,
        func: Callable[[datetime, datetime, int, int, float, int, float], NoReturn] = None,
    ) -> float:
        """
                                       ts_prev,  ts_curr, mag_prev,  ts_delta,   area
                                                                mag_curr,   mag_delta
        This is basically geometry math.  We treat the magnitude graph
        as a set of blocks to calculate the area.
        edges must already be present"""
        area: float = 0.0
        for ts_prev_mag_prev, ts_curr_mag_curr, ts_next_mag_next in iterate_chain(mag_timeline):
            if ts_prev_mag_prev is None:
                # first
                continue
            (ts_prev, mag_prev), (ts_curr, mag_curr), _ = (
                ts_prev_mag_prev,
                ts_curr_mag_curr,
                ts_next_mag_next,
            )
            ts_delta: Union[pd.Timedelta, timedelta] = ts_curr - ts_prev
            if type(ts_delta) in (pd.Timedelta, timedelta):
                ts_delta_float = ts_delta.total_seconds()
            else:
                ts_delta_float = ts_delta
            mag_delta = mag_curr - mag_prev
            # once you detect the block which the magnitude has not changed,
            # calculate the area
            if func:
                func(
                    ts_prev,
                    ts_curr,
                    mag_prev,
                    mag_curr,
                    ts_delta_float,
                    mag_delta,
                    area,
                )
            if mag_delta == 0:
                curr_area = ts_delta_float * mag_curr
                area += curr_area
        return area

    def find_event_dependencies(self, event_lst, dependency_functions, outer_func=None) -> List[Dependency]:
        """search through all events and create a list of dependencies.
        This function will loop through all events for all events O(N**2),
        and will create dependencies in both directions, if they exist.
        """
        # loop through all events
        dependency_lst = []
        for oidx, oevent_dct in enumerate(event_lst):
            if outer_func:
                outer_func(oidx, oevent_dct)
            for iidx, ievent_dct in enumerate(event_lst):  # event_lst[oidx:]: #disabled due to implicit ordering
                # lets not add ourself
                if oidx != iidx:
                    deps = []
                    for func in dependency_functions:
                        dependant = func(oevent_dct, ievent_dct)
                        deps.append(dependant)
                    if all(deps):
                        dependency_lst.append(Dependency(oevent_dct, ievent_dct))
        return dependency_lst

    def merge_event_dependencies(self, dependency_lst) -> Dict:
        """take all given dependencies and join them together.
        Since dependencies are not guaranteed ordered, we will have to join knowing this.
        """
        groupings = defaultdict(list)
        for dependency in dependency_lst:
            a_pk = dependency.a_pk
            b_pk = dependency.b_pk
            groupings[a_pk].append(b_pk)
        return groupings

    def generate_dependency_graph(self, node_lst, dependency_lst, filename):
        """generate a dot graph of the dependencies."""
        import graphviz as gv

        g = gv.Digraph(format="svg")
        for node in node_lst:
            g.node(node)
        for dependency in dependency_lst:
            g.edge(dependency.a_pk, dependency.b_pk)
        g.render("img/%s" % filename)


class Event_Handler:
    """This creates local uniqueness for pk's with in a system.
    This can be used to create dependencies from one event to another using a Guaranteed unique with the class, pk.
    It also provides a lookup from pk -> event.
    """

    def __init__(self, unique_func=None):
        self.pk_lookup = {}
        self.nodes = []
        if unique_func is None:
            self.unique_func = self._get_uuid
            self.max_loop_count = 10  # how could we ever hit this!
        else:
            self.unique_func = unique_func
            self.max_loop_count = 1

    def _get_uuid(self, initial):
        return uuid.uuid4().hex

    def pk_func(self, initial):
        """guarantee uniqueness in uuids

        loop while there is no uuid
            get a uuid4
            if not in self.pk_lookup.keys(), leave it set
            else: set it to none to get another.
        """
        pk = None
        max_loop_count = self.max_loop_count
        while not pk and max_loop_count > 0:
            pk = self.unique_func(initial)
            if pk in self.pk_lookup.keys():
                # found a duplicate
                print("Warning:found a duplicate pk of %s for %s" % (pk, initial))
                pk = None
            max_loop_count -= 1
        if pk is None:
            raise Exception("Could not generate unique key for %s" % initial)
        return pk

    def _associate_pk_with_event(self, pk, event_dct):
        """associate the pk with the event.  This has side effects and
        really isn't needed due to the pk is pulled from get_uuid and already
        checks this.  TODO: find a way to tie the two together to remove
        this function."""
        try:
            self.pk_lookup[pk]
        except KeyError:
            # good no duplicates, whew!
            self.pk_lookup[pk] = event_dct
            self.nodes.append(pk)
        else:
            # duplicate, exit now!
            raise DuplicateWarning("Duplicated pk %s found, from e(%s) to e(%s)" % (pk, self.pk_lookup[pk], event_dct))

    def init_event(self, initial):
        """we need a consistent way to create an event"""
        assert "pk" not in initial, "setting pk is not allowed"
        pk = self.pk_func(initial)
        event_dct = OrderedDict(pk=pk, **initial)
        self._associate_pk_with_event(pk, event_dct)
        return event_dct

    def copy_event(self, event_dct):
        """when we copy an event, we need to generate a new key"""
        # we are not allowed to have pk set, remove it.
        initial = {key: value for key, value in event_dct.items() if key != "pk"}
        return self.init_event(initial)

    def get_event(
        self,
        event_type_name: str,
        bitmask: EnhancedBitmaskLike,
        time_start: datetime,
        time_end: datetime,
    ) -> TLEvent:
        initial = dict(
            event_type_name=event_type_name,
            bitmask=bitmask,
            time_start=time_start,
            time_end=time_end,
        )
        pk = self.pk_func(initial)
        event_dct: TLEvent = TLEvent(
            pk=pk,
            event_type_name=event_type_name,
            bitmask=bitmask,
            time_start=time_start,
            time_end=time_end,
        )
        self._associate_pk_with_event(pk, event_dct)
        return event_dct


class Dependency_Funcs:
    @classmethod
    def dep_time(cls, event_a, event_b):
        intersect = False
        a_ts = event_a["time_start"]
        a_te = event_a["time_end"]
        a_ir = IntervalRange(a_ts, a_te)
        b_ts = event_b["time_start"]
        b_te = event_b["time_end"]
        b_ir = IntervalRange(b_ts, b_te)
        # check if either b is in range a
        if a_ir.in_range(b_ts) or a_ir.in_range(b_te):
            intersect = True
        if b_ir.in_range(a_ts) or b_ir.in_range(a_te):
            intersect = True
        return intersect

    @classmethod
    def dep_space_all(cls, event_a, event_b):
        intersect = False
        a_bitmask = event_a["bitmask"]
        b_bitmask = event_b["bitmask"]
        if a_bitmask.logical_xnor(a_bitmask, b_bitmask).all() == True:
            intersect = True
        return intersect
